class PyliteTypes:
    SupportedTypes = [
        int,
        float,  # Numeric
        str,  # String
        bool,  # Boolean
        list,  # List
        dict,  # Dictionary
    ]


class email:
    def __init__(self, email):
        self.email = email
